@@ -6,14 +6,16 @@
 # This software may be modified and distributed under the terms
 # of the BSD license.  See the LICENSE file for details.
 
-import datetime
+import warnings
 
-from aniso8601.date import parse_date
-from aniso8601.exceptions import ISOFormatError, RelativeValueError
-from aniso8601.time import parse_time
 from aniso8601 import compat
+from aniso8601.builder import (TupleBuilder, PythonTimeBuilder,
+                               RelativeTimeBuilder)
+from aniso8601.date import parse_date
+from aniso8601.exceptions import ISOFormatError, NegativeDurationError
+from aniso8601.time import parse_time
 
-def parse_duration(isodurationstr, relative=False):
+def parse_duration(isodurationstr, relative=False, builder=PythonTimeBuilder):
     #Given a string representing an ISO 8601 duration, return a
     #datetime.timedelta (or dateutil.relativedelta.relativedelta
     #if relative=True) that matches the given duration. Valid formats are:
@@ -24,23 +26,38 @@ def parse_duration(isodurationstr, relative=False):
     if isodurationstr[0] != 'P':
         raise ISOFormatError('ISO 8601 duration must start with a P.')
 
-    #If Y, M, D, H, S, or W are in the string, assume it is a specified duration
-    if _has_any_component(isodurationstr, ['Y', 'M', 'D', 'H', 'S', 'W']) is True:
-        return _parse_duration_prescribed(isodurationstr, relative)
+    if relative is True:
+        warnings.warn('relative kwarg is deprecated',
+                      DeprecationWarning, stacklevel=2)
 
-    return _parse_duration_combined(isodurationstr, relative)
+        builder = RelativeTimeBuilder
 
-def _parse_duration_prescribed(durationstr, relative):
+    #If Y, M, D, H, S, or W are in the string,
+    #assume it is a specified duration
+    if _has_any_component(isodurationstr,
+                          ['Y', 'M', 'D', 'H', 'S', 'W']) is True:
+        return _parse_duration_prescribed(isodurationstr, builder)
+
+    return _parse_duration_combined(isodurationstr, builder)
+
+def _parse_duration_prescribed(durationstr, builder):
     #durationstr can be of the form PnYnMnDTnHnMnS or PnW
 
+    #Don't allow negative elements
+    #https://bitbucket.org/nielsenb/aniso8601/issues/20/negative-duration
+    if durationstr.find('-') != -1:
+        raise NegativeDurationError('ISO 8601 durations must be positive.')
+
     #Make sure the end character is valid
     #https://bitbucket.org/nielsenb/aniso8601/issues/9/durations-with-trailing-garbage-are-parsed
     if durationstr[-1] not in ['Y', 'M', 'D', 'H', 'S', 'W']:
-        raise ISOFormatError('ISO 8601 duration must end with a valid character.')
+        raise ISOFormatError('ISO 8601 duration must end with a valid '
+                             'character.')
 
     #Make sure only the lowest order element has decimal precision
     if durationstr.count('.') > 1:
-        raise ISOFormatError('ISO 8601 allows only lowest order element to have a decimal fraction.')
+        raise ISOFormatError('ISO 8601 allows only lowest order element to '
+                             'have a decimal fraction.')
     elif durationstr.count('.') == 1:
         #There should only ever be 1 letter after a decimal if there is more
         #then one, the string is invalid
@@ -50,151 +67,141 @@ def _parse_duration_prescribed(durationstr, relative):
             if character.isalpha() is True:
                 lettercount += 1
 
-            if lettercount > 1:
-                raise ISOFormatError('ISO 8601 duration must end with a single valid character.')
+                if lettercount > 1:
+                    raise ISOFormatError('ISO 8601 duration must end with '
+                                         'a single valid character.')
 
     #Do not allow W in combination with other designators
     #https://bitbucket.org/nielsenb/aniso8601/issues/2/week-designators-should-not-be-combinable
-    if durationstr.find('W') != -1 and _has_any_component(durationstr, ['Y', 'M', 'D', 'H', 'S']) is True:
-        raise ISOFormatError('ISO 8601 week designators may not be combined with other time designators.')
+    if (durationstr.find('W') != -1
+            and _has_any_component(durationstr,
+                                   ['Y', 'M', 'D', 'H', 'S']) is True):
+        raise ISOFormatError('ISO 8601 week designators may not be combined '
+                             'with other time designators.')
 
     #Parse the elements of the duration
     if durationstr.find('T') == -1:
-        years, months, weeks, days, hours, minutes, seconds = _parse_duration_prescribed_notime(durationstr)
-    else:
-        years, months, weeks, days, hours, minutes, seconds = _parse_duration_prescribed_time(durationstr)
+        return _parse_duration_prescribed_notime(durationstr, builder)
 
-    if relative is True:
-        try:
-            import dateutil.relativedelta
+    return _parse_duration_prescribed_time(durationstr, builder)
 
-            if int(years) != years or int(months) != months:
-                #https://github.com/dateutil/dateutil/issues/40
-                raise RelativeValueError('Fractional months and years are not defined for relative intervals.')
-
-            return dateutil.relativedelta.relativedelta(years=int(years), months=int(months), weeks=weeks, days=days, hours=hours, minutes=minutes, seconds=seconds)
-        except ImportError:
-            raise RuntimeError('dateutil must be installed for relative duration support.')
-
-    #Note that weeks can be handled without conversion to days
-    totaldays = years * 365 + months * 30 + days
-
-    return datetime.timedelta(weeks=weeks, days=totaldays, hours=hours, minutes=minutes, seconds=seconds)
-
-def _parse_duration_prescribed_notime(durationstr):
+def _parse_duration_prescribed_notime(durationstr, builder):
     #durationstr can be of the form PnYnMnD or PnW
 
+    #Don't allow negative elements
+    #https://bitbucket.org/nielsenb/aniso8601/issues/20/negative-duration
+    if durationstr.find('-') != -1:
+        raise NegativeDurationError('ISO 8601 durations must be positive.')
+
     #Make sure no time portion is included
     #https://bitbucket.org/nielsenb/aniso8601/issues/7/durations-with-time-components-before-t
     if _has_any_component(durationstr, ['H', 'S']):
-        raise ISOFormatError('ISO 8601 time components not allowed in duration without prescribed time.')
+        raise ISOFormatError('ISO 8601 time components not allowed in duration '
+                             'without prescribed time.')
 
-    if _component_order_correct(durationstr, ['P', 'Y', 'M', 'D', 'W']) is False:
-        raise ISOFormatError('ISO 8601 duration components must be in the correct order.')
+    if _component_order_correct(durationstr,
+                                ['P', 'Y', 'M', 'D', 'W']) is False:
+        raise ISOFormatError('ISO 8601 duration components must be in the '
+                             'correct order.')
 
     if durationstr.find('Y') != -1:
-        years = _parse_duration_element(durationstr, 'Y')
+        yearstr = _parse_duration_element(durationstr, 'Y')
     else:
-        years = 0
+        yearstr = None
 
     if durationstr.find('M') != -1:
-        months = _parse_duration_element(durationstr, 'M')
+        monthstr = _parse_duration_element(durationstr, 'M')
     else:
-        months = 0
+        monthstr = None
 
     if durationstr.find('W') != -1:
-        weeks = _parse_duration_element(durationstr, 'W')
+        weekstr = _parse_duration_element(durationstr, 'W')
     else:
-        weeks = 0
+        weekstr = None
 
     if durationstr.find('D') != -1:
-        days = _parse_duration_element(durationstr, 'D')
+        daystr = _parse_duration_element(durationstr, 'D')
     else:
-        days = 0
+        daystr = None
 
-    #No hours, minutes or seconds
-    hours = 0
-    minutes = 0
-    seconds = 0
+    return builder.build_duration(PnY=yearstr, PnM=monthstr,
+                                  PnW=weekstr, PnD=daystr)
 
-    return (years, months, weeks, days, hours, minutes, seconds)
-
-def _parse_duration_prescribed_time(durationstr):
+def _parse_duration_prescribed_time(durationstr, builder):
     #durationstr can be of the form PnYnMnDTnHnMnS
 
+    #Don't allow negative elements
+    #https://bitbucket.org/nielsenb/aniso8601/issues/20/negative-duration
+    if durationstr.find('-') != -1:
+        raise NegativeDurationError('ISO 8601 durations must be positive.')
+
     firsthalf = durationstr[:durationstr.find('T')]
     secondhalf = durationstr[durationstr.find('T'):]
 
     #Make sure no time portion is included in the date half
     #https://bitbucket.org/nielsenb/aniso8601/issues/7/durations-with-time-components-before-t
     if _has_any_component(firsthalf, ['H', 'S']):
-        raise ISOFormatError('ISO 8601 time components not allowed in date portion of duration.')
+        raise ISOFormatError('ISO 8601 time components not allowed in date '
+                             'portion of duration.')
 
     if _component_order_correct(firsthalf, ['P', 'Y', 'M', 'D', 'W']) is False:
-        raise ISOFormatError('ISO 8601 duration components must be in the correct order.')
+        raise ISOFormatError('ISO 8601 duration components must be in the '
+                             'correct order.')
 
     #Make sure no date component is included in the time half
     if _has_any_component(secondhalf, ['Y', 'D']):
-        raise ISOFormatError('ISO 8601 time components not allowed in date portion of duration.')
+        raise ISOFormatError('ISO 8601 time components not allowed in date '
+                             'portion of duration.')
 
     if _component_order_correct(secondhalf, ['T', 'H', 'M', 'S']) is False:
-        raise ISOFormatError('ISO 8601 time components in duration must be in the correct order.')
+        raise ISOFormatError('ISO 8601 time components in duration must be in '
+                             'the correct order.')
 
     if firsthalf.find('Y') != -1:
-        years = _parse_duration_element(firsthalf, 'Y')
+        yearstr = _parse_duration_element(firsthalf, 'Y')
     else:
-        years = 0
+        yearstr = None
 
     if firsthalf.find('M') != -1:
-        months = _parse_duration_element(firsthalf, 'M')
+        monthstr = _parse_duration_element(firsthalf, 'M')
     else:
-        months = 0
+        monthstr = None
 
     if firsthalf.find('D') != -1:
-        days = _parse_duration_element(firsthalf, 'D')
+        daystr = _parse_duration_element(firsthalf, 'D')
     else:
-        days = 0
+        daystr = None
 
     if secondhalf.find('H') != -1:
-        hours = _parse_duration_element(secondhalf, 'H')
+        hourstr = _parse_duration_element(secondhalf, 'H')
     else:
-        hours = 0
+        hourstr = None
 
     if secondhalf.find('M') != -1:
-        minutes = _parse_duration_element(secondhalf, 'M')
+        minutestr = _parse_duration_element(secondhalf, 'M')
     else:
-        minutes = 0
+        minutestr = None
 
     if secondhalf.find('S') != -1:
-        seconds = _parse_duration_element(secondhalf, 'S')
+        secondstr = _parse_duration_element(secondhalf, 'S')
     else:
-        seconds = 0
+        secondstr = None
 
-    #Weeks can't be included
-    weeks = 0
+    return builder.build_duration(PnY=yearstr, PnM=monthstr, PnD=daystr,
+                                  TnH=hourstr, TnM=minutestr, TnS=secondstr)
 
-    return (years, months, weeks, days, hours, minutes, seconds)
-
-def _parse_duration_combined(durationstr, relative):
+def _parse_duration_combined(durationstr, builder):
     #Period of the form P<date>T<time>
 
     #Split the string in to its component parts
     datepart, timepart = durationstr[1:].split('T') #We skip the 'P'
 
-    datevalue = parse_date(datepart)
-    timevalue = parse_time(timepart)
+    datevalue = parse_date(datepart, builder=TupleBuilder)
+    timevalue = parse_time(timepart, builder=TupleBuilder)
 
-    if relative is True:
-        try:
-            import dateutil.relativedelta
-
-            return dateutil.relativedelta.relativedelta(years=datevalue.year, months=datevalue.month, days=datevalue.day, hours=timevalue.hour, minutes=timevalue.minute, seconds=timevalue.second, microseconds=timevalue.microsecond)
-        except ImportError:
-            raise RuntimeError('dateutil must be installed for relative duration support.')
-    else:
-        totaldays = datevalue.year * 365 + datevalue.month * 30 + datevalue.day
-
-        return datetime.timedelta(days=totaldays, hours=timevalue.hour, minutes=timevalue.minute, seconds=timevalue.second, microseconds=timevalue.microsecond)
+    return builder.build_duration(PnY=datevalue[0], PnM=datevalue[1],
+                                  PnD=datevalue[2], TnH=timevalue[0],
+                                  TnM=timevalue[1], TnS=timevalue[2])
 
 def _parse_duration_element(durationstr, elementstr):
     #Extracts the specified portion of a duration, for instance, given:
@@ -220,16 +227,7 @@ def _parse_duration_element(durationstr, elementstr):
         #Replace the comma with a 'full-stop'
         durationstr = durationstr.replace(',', '.')
 
-    if elementstr == 'S':
-        #We truncate seconds to avoid precision issues with microseconds
-        #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
-        if '.' in durationstr[durationstartindex:durationendindex]:
-            stopindex = durationstr.index('.')
-
-            if durationendindex - stopindex > 7:
-                durationendindex = stopindex + 7
-
-    return float(durationstr[durationstartindex:durationendindex])
+    return durationstr[durationstartindex:durationendindex]
 
 def _has_any_component(durationstr, components):
     #Given a duration string, and a list of components, returns True
